# Copyright 2014-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""A fake SSLContext implementation."""

import ssl as _ssl
import sys as _sys

# PROTOCOL_TLS_CLIENT is Python 3.6+
PROTOCOL_SSLv23 = getattr(_ssl, "PROTOCOL_TLS_CLIENT", _ssl.PROTOCOL_SSLv23)
# Python 2.7.9+
OP_NO_SSLv2 = getattr(_ssl, "OP_NO_SSLv2", 0)
# Python 2.7.9+
OP_NO_SSLv3 = getattr(_ssl, "OP_NO_SSLv3", 0)
# Python 2.7.9+, OpenSSL 1.0.0+
OP_NO_COMPRESSION = getattr(_ssl, "OP_NO_COMPRESSION", 0)
# Python 3.7+, OpenSSL 1.1.0h+
OP_NO_RENEGOTIATION = getattr(_ssl, "OP_NO_RENEGOTIATION", 0)

# Python 2.7.9+
HAS_SNI = getattr(_ssl, "HAS_SNI", False)
IS_PYOPENSSL = False

# Base Exception class
SSLError = _ssl.SSLError

try:
    # CPython 2.7.9+
    from ssl import SSLContext
    if hasattr(_ssl, "VERIFY_CRL_CHECK_LEAF"):
        from ssl import VERIFY_CRL_CHECK_LEAF
    # Python 3.7 uses OpenSSL's hostname matching implementation
    # making it the obvious version to start using SSLConext.check_hostname.
    # Python 3.6 might have been a good version, but it suffers
    # from https://bugs.python.org/issue32185.
    # We'll use our bundled match_hostname for older Python
    # versions, which also supports IP address matching
    # with Python < 3.5.
    CHECK_HOSTNAME_SAFE = _sys.version_info[:2] >= (3, 7)
except ImportError:
    from pymongo.errors import ConfigurationError

    class SSLContext(object):
        """A fake SSLContext.

        This implements an API similar to ssl.SSLContext from python 3.2
        but does not implement methods or properties that would be
        incompatible with ssl.wrap_socket from python 2.7 < 2.7.9.

        You must pass protocol which must be one of the PROTOCOL_* constants
        defined in the ssl module. ssl.PROTOCOL_SSLv23 is recommended for maximum
        interoperability.
        """

        __slots__ = ('_cafile', '_certfile',
                     '_keyfile', '_protocol', '_verify_mode')

        def __init__(self, protocol):
            self._cafile = None
            self._certfile = None
            self._keyfile = None
            self._protocol = protocol
            self._verify_mode = _ssl.CERT_NONE

        @property
        def protocol(self):
            """The protocol version chosen when constructing the context.
            This attribute is read-only.
            """
            return self._protocol

        def __get_verify_mode(self):
            """Whether to try to verify other peers' certificates and how to
            behave if verification fails. This attribute must be one of
            ssl.CERT_NONE, ssl.CERT_OPTIONAL or ssl.CERT_REQUIRED.
            """
            return self._verify_mode

        def __set_verify_mode(self, value):
            """Setter for verify_mode."""
            self._verify_mode = value

        verify_mode = property(__get_verify_mode, __set_verify_mode)

        def load_cert_chain(self, certfile, keyfile=None, password=None):
            """Load a private key and the corresponding certificate. The certfile
            string must be the path to a single file in PEM format containing the
            certificate as well as any number of CA certificates needed to
            establish the certificate's authenticity. The keyfile string, if
            present, must point to a file containing the private key. Otherwise
            the private key will be taken from certfile as well.
            """
            if password is not None:
                raise ConfigurationError(
                    "Support for ssl_pem_passphrase requires "
                    "python 2.7.9+ (pypy 2.5.1+), python 3 or "
                    "PyOpenSSL")
            self._certfile = certfile
            self._keyfile = keyfile

        def load_verify_locations(self, cafile=None, dummy=None):
            """Load a set of "certification authority"(CA) certificates used to
            validate other peers' certificates when `~verify_mode` is other than
            ssl.CERT_NONE.
            """
            self._cafile = cafile

        def wrap_socket(self, sock, server_side=False,
                        do_handshake_on_connect=True,
                        suppress_ragged_eofs=True, dummy=None):
            """Wrap an existing Python socket sock and return an ssl.SSLSocket
            object.
            """
            return _ssl.wrap_socket(sock, keyfile=self._keyfile,
                                    certfile=self._certfile,
                                    server_side=server_side,
                                    cert_reqs=self._verify_mode,
                                    ssl_version=self._protocol,
                                    ca_certs=self._cafile,
                                    do_handshake_on_connect=do_handshake_on_connect,
                                    suppress_ragged_eofs=suppress_ragged_eofs)



import os
import platform
import re
import sys
import warnings


if sys.version_info[:2] < (2, 7):
    raise RuntimeError("Python version >= 2.7 required.")


# Hack to silence atexit traceback in some Python versions
try:
    import multiprocessing
except ImportError:
    pass

# Don't force people to install setuptools unless
# we have to.
try:
    from setuptools import setup, __version__ as _setuptools_version
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, __version__ as _setuptools_version

from distutils.cmd import Command
from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsOptionError
from distutils.errors import DistutilsPlatformError, DistutilsExecError
from distutils.core import Extension

_HAVE_SPHINX = True
try:
    from sphinx.cmd import build as sphinx
except ImportError:
    try:
        import sphinx
    except ImportError:
        _HAVE_SPHINX = False

version = "3.11.1"

f = open("README.rst")
try:
    try:
        readme_content = f.read()
    except:
        readme_content = ""
finally:
    f.close()

# PYTHON-654 - Clang doesn't support -mno-fused-madd but the pythons Apple
# ships are built with it. This is a problem starting with Xcode 5.1
# since clang 3.4 errors out when it encounters unrecognized compiler
# flags. This hack removes -mno-fused-madd from the CFLAGS automatically
# generated by distutils for Apple provided pythons, allowing C extension
# builds to complete without error. The inspiration comes from older
# versions of distutils.sysconfig.get_config_vars.
if sys.platform == 'darwin' and 'clang' in platform.python_compiler().lower():
    from distutils.sysconfig import get_config_vars
    res = get_config_vars()
    for key in ('CFLAGS', 'PY_CFLAGS'):
        if key in res:
            flags = res[key]
            flags = re.sub('-mno-fused-madd', '', flags)
            res[key] = flags


class test(Command):
    description = "run the tests"

    user_options = [
        ("test-module=", "m", "Discover tests in specified module"),
        ("test-suite=", "s",
         "Test suite to run (e.g. 'some_module.test_suite')"),
        ("failfast", "f", "Stop running tests on first failure or error"),
        ("xunit-output=", "x",
         "Generate a results directory with XUnit XML format")
    ]

    def initialize_options(self):
        self.test_module = None
        self.test_suite = None
        self.failfast = False
        self.xunit_output = None

    def finalize_options(self):
        if self.test_suite is None and self.test_module is None:
            self.test_module = 'test'
        elif self.test_module is not None and self.test_suite is not None:
            raise DistutilsOptionError(
                "You may specify a module or suite, but not both"
            )

    def run(self):
        # Installing required packages, running egg_info and build_ext are
        # part of normal operation for setuptools.command.test.test
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(
                self.distribution.install_requires)
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)
        if self.xunit_output:
            self.distribution.fetch_build_eggs(["unittest-xml-reporting"])
        self.run_command('egg_info')
        build_ext_cmd = self.reinitialize_command('build_ext')
        build_ext_cmd.inplace = 1
        self.run_command('build_ext')

        # Construct a TextTestRunner directly from the unittest imported from
        # test, which creates a TestResult that supports the 'addSkip' method.
        # setuptools will by default create a TextTestRunner that uses the old
        # TestResult class.
        from test import unittest, PymongoTestRunner, test_cases
        if self.test_suite is None:
            all_tests = unittest.defaultTestLoader.discover(self.test_module)
            suite = unittest.TestSuite()
            suite.addTests(sorted(test_cases(all_tests),
                                  key=lambda x: x.__module__))
        else:
            suite = unittest.defaultTestLoader.loadTestsFromName(
                self.test_suite)
        if self.xunit_output:
            from test import PymongoXMLTestRunner
            runner = PymongoXMLTestRunner(verbosity=2, failfast=self.failfast,
                                          output=self.xunit_output)
        else:
            runner = PymongoTestRunner(verbosity=2, failfast=self.failfast)
        result = runner.run(suite)
        sys.exit(not result.wasSuccessful())


class doc(Command):

    description = "generate or test documentation"

    user_options = [("test", "t",
                     "run doctests instead of generating documentation")]

    boolean_options = ["test"]

    def initialize_options(self):
        self.test = False

    def finalize_options(self):
        pass

    def run(self):

        if not _HAVE_SPHINX:
            raise RuntimeError(
                "You must install Sphinx to build or test the documentation.")

        if sys.version_info[0] >= 3:
            import doctest
            from doctest import OutputChecker as _OutputChecker

            # Match u or U (possibly followed by r or R), removing it.
            # r/R can follow u/U but not precede it. Don't match the
            # single character string 'u' or 'U'.
            _u_literal_re = re.compile(
                r"(\W|^)(?<![\'\"])[uU]([rR]?[\'\"])", re.UNICODE)
             # Match b or B (possibly followed by r or R), removing.
             # r/R can follow b/B but not precede it. Don't match the
             # single character string 'b' or 'B'.
            _b_literal_re = re.compile(
                r"(\W|^)(?<![\'\"])[bB]([rR]?[\'\"])", re.UNICODE)

            class _StringPrefixFixer(_OutputChecker):

                def check_output(self, want, got, optionflags):
                    # The docstrings are written with python 2.x in mind.
                    # To make the doctests pass in python 3 we have to
                    # strip the 'u' prefix from the expected results. The
                    # actual results won't have that prefix.
                    want = re.sub(_u_literal_re, r'\1\2', want)
                    # We also have to strip the 'b' prefix from the actual
                    # results since python 2.x expected results won't have
                    # that prefix.
                    got = re.sub(_b_literal_re, r'\1\2', got)
                    return super(
                        _StringPrefixFixer, self).check_output(
                            want, got, optionflags)

                def output_difference(self, example, got, optionflags):
                    example.want = re.sub(_u_literal_re, r'\1\2', example.want)
                    got = re.sub(_b_literal_re, r'\1\2', got)
                    return super(
                        _StringPrefixFixer, self).output_difference(
                            example, got, optionflags)

            doctest.OutputChecker = _StringPrefixFixer

        if self.test:
            path = os.path.join(
                os.path.abspath('.'), "doc", "_build", "doctest")
            mode = "doctest"
        else:
            path = os.path.join(
                os.path.abspath('.'), "doc", "_build", version)
            mode = "html"

            try:
                os.makedirs(path)
            except:
                pass

        sphinx_args = ["-E", "-b", mode, "doc", path]

        # sphinx.main calls sys.exit when sphinx.build_main exists.
        # Call build_main directly so we can check status and print
        # the full path to the built docs.
        if hasattr(sphinx, 'build_main'):
            status = sphinx.build_main(sphinx_args)
        else:
            status = sphinx.main(sphinx_args)

        if status:
            raise RuntimeError("documentation step '%s' failed" % (mode,))

        sys.stdout.write("\nDocumentation step '%s' performed, results here:\n"
                         "   %s/\n" % (mode, path))


if sys.platform == 'win32':
    # distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    build_errors = (CCompilerError, DistutilsExecError,
                    DistutilsPlatformError, IOError)
else:
    build_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)


class custom_build_ext(build_ext):
    """Allow C extension building to fail.

    The C extension speeds up BSON encoding, but is not essential.
    """

    warning_message = """
********************************************************************
WARNING: %s could not
be compiled. No C extensions are essential for PyMongo to run,
although they do result in significant speed improvements.
%s

Please see the installation docs for solutions to build issues:

https://pymongo.readthedocs.io/en/stable/installation.html

Here are some hints for popular operating systems:

If you are seeing this message on Linux you probably need to
install GCC and/or the Python development package for your
version of Python.

Debian and Ubuntu users should issue the following command:

    $ sudo apt-get install build-essential python-dev

Users of Red Hat based distributions (RHEL, CentOS, Amazon Linux,
Oracle Linux, Fedora, etc.) should issue the following command:

    $ sudo yum install gcc python-devel

If you are seeing this message on Microsoft Windows please install
PyMongo using pip. Modern versions of pip will install PyMongo
from binary wheels available on pypi. If you must install from
source read the documentation here:

https://pymongo.readthedocs.io/en/stable/installation.html#installing-from-source-on-windows

If you are seeing this message on macOS / OSX please install PyMongo
using pip. Modern versions of pip will install PyMongo from binary
wheels available on pypi. If wheels are not available for your version
of macOS / OSX, or you must install from source read the documentation
here:

https://pymongo.readthedocs.io/en/stable/installation.html#osx
********************************************************************
"""

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            e = sys.exc_info()[1]
            sys.stdout.write('%s\n' % str(e))
            warnings.warn(self.warning_message % ("Extension modules",
                                                  "There was an issue with "
                                                  "your platform configuration"
                                                  " - see above."))

    def build_extension(self, ext):
        name = ext.name
        try:
            build_ext.build_extension(self, ext)
        except build_errors:
            e = sys.exc_info()[1]
            sys.stdout.write('%s\n' % str(e))
            warnings.warn(self.warning_message % ("The %s extension "
                                                  "module" % (name,),
                                                  "The output above "
                                                  "this warning shows how "
                                                  "the compilation "
                                                  "failed."))

ext_modules = [Extension('bson._cbson',
                         include_dirs=['bson'],
                         sources=['bson/_cbsonmodule.c',
                                  'bson/time64.c',
                                  'bson/buffer.c',
                                  'bson/encoding_helpers.c']),
               Extension('pymongo._cmessage',
                         include_dirs=['bson'],
                         sources=['pymongo/_cmessagemodule.c',
                                  'bson/buffer.c'])]

# PyOpenSSL 17.0.0 introduced support for OCSP. 17.1.0 introduced
# a related feature we need. 17.2.0 fixes a bug
# in set_default_verify_paths we should really avoid.
# service_identity 18.1.0 introduced support for IP addr matching.
pyopenssl_reqs = ["pyopenssl>=17.2.0", "requests<3.0.0", "service_identity>=18.1.0"]

extras_require = {
    'encryption': ['pymongocrypt<2.0.0'],
    'ocsp': pyopenssl_reqs,
    'snappy': ['python-snappy'],
    'tls': [],
    'zstd': ['zstandard'],
    'aws': ['pymongo-auth-aws<2.0.0'],
}

# https://jira.mongodb.org/browse/PYTHON-2117
# Environment marker support didn't settle down until version 20.10
# https://setuptools.readthedocs.io/en/latest/history.html#v20-10-0
_use_env_markers = tuple(map(int, _setuptools_version.split('.')[:2])) > (20, 9)

# TLS and DNS extras
# We install PyOpenSSL and service_identity for Python < 2.7.9 to
# get support for SNI, which is required to connection to Altas
# free and shared tier.
if sys.version_info[0] == 2:
    if _use_env_markers:
        # For building wheels on Python versions >= 2.7.9
        for req in pyopenssl_reqs:
            extras_require['tls'].append(
                "%s ; python_full_version < '2.7.9'" % (req,))
        if sys.platform == 'win32':
            extras_require['tls'].append(
                "wincertstore>=0.2 ; python_full_version < '2.7.9'")
        else:
            extras_require['tls'].append(
                "certifi ; python_full_version < '2.7.9'")
    elif sys.version_info < (2, 7, 9):
        # For installing from source or egg files on Python versions
        # older than 2.7.9, or systems that have setuptools versions
        # older than 20.10.
        extras_require['tls'].extend(pyopenssl_reqs)
        if sys.platform == 'win32':
            extras_require['tls'].append("wincertstore>=0.2")
        else:
            extras_require['tls'].append("certifi")
    extras_require.update({'srv': ["dnspython>=1.16.0,<1.17.0"]})
    extras_require.update({'tls': ["ipaddress"]})
else:
    extras_require.update({'srv': ["dnspython>=1.16.0,<2.0.0"]})

# GSSAPI extras
if sys.platform == 'win32':
    extras_require['gssapi'] = ["winkerberos>=0.5.0"]
else:
    extras_require['gssapi'] = ["pykerberos"]

extra_opts = {
    "packages": ["bson", "pymongo", "gridfs"]
}

if "--no_ext" in sys.argv:
    sys.argv.remove("--no_ext")
elif (sys.platform.startswith("java") or
      sys.platform == "cli" or
      "PyPy" in sys.version):
    sys.stdout.write("""
*****************************************************\n
The optional C extensions are currently not supported\n
by this python implementation.\n
*****************************************************\n
""")
else:
    extra_opts['ext_modules'] = ext_modules

setup(
    name="pymongo",
    version=version,
    description="Python driver for MongoDB <http://www.mongodb.org>",
    long_description=readme_content,
    author="Mike Dirolf",
    author_email="mongodb-user@googlegroups.com",
    maintainer="Bernie Hackett",
    maintainer_email="bernie@mongodb.com",
    url="http://github.com/mongodb/mongo-python-driver",
    keywords=["mongo", "mongodb", "pymongo", "gridfs", "bson"],
    install_requires=[],
    license="Apache License, Version 2.0",
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database"],
    cmdclass={"build_ext": custom_build_ext,
              "doc": doc,
              "test": test},
    extras_require=extras_require,
    **extra_opts
)



# Copyright 2016-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the collation module."""

import functools
import warnings

from pymongo.collation import (
    Collation,
    CollationCaseFirst, CollationStrength, CollationAlternate,
    CollationMaxVariable)
from pymongo.errors import ConfigurationError
from pymongo.operations import (DeleteMany, DeleteOne, IndexModel, ReplaceOne,
                                UpdateMany, UpdateOne)
from pymongo.write_concern import WriteConcern
from test import unittest, client_context
from test.utils import EventListener, ignore_deprecations, rs_or_single_client


class TestCollationObject(unittest.TestCase):

    def test_constructor(self):
        self.assertRaises(TypeError, Collation, locale=42)
        # Fill in a locale to test the other options.
        _Collation = functools.partial(Collation, 'en_US')
        # No error.
        _Collation(caseFirst=CollationCaseFirst.UPPER)
        self.assertRaises(TypeError, _Collation, caseLevel='true')
        self.assertRaises(ValueError, _Collation, strength='six')
        self.assertRaises(TypeError, _Collation,
                          numericOrdering='true')
        self.assertRaises(TypeError, _Collation, alternate=5)
        self.assertRaises(TypeError, _Collation, maxVariable=2)
        self.assertRaises(TypeError, _Collation, normalization='false')
        self.assertRaises(TypeError, _Collation, backwards='true')

        # No errors.
        Collation('en_US', future_option='bar', another_option=42)
        collation = Collation(
            'en_US',
            caseLevel=True,
            caseFirst=CollationCaseFirst.UPPER,
            strength=CollationStrength.QUATERNARY,
            numericOrdering=True,
            alternate=CollationAlternate.SHIFTED,
            maxVariable=CollationMaxVariable.SPACE,
            normalization=True,
            backwards=True)

        self.assertEqual({
            'locale': 'en_US',
            'caseLevel': True,
            'caseFirst': 'upper',
            'strength': 4,
            'numericOrdering': True,
            'alternate': 'shifted',
            'maxVariable': 'space',
            'normalization': True,
            'backwards': True
        }, collation.document)

        self.assertEqual({
            'locale': 'en_US',
            'backwards': True
        }, Collation('en_US', backwards=True).document)


def raisesConfigurationErrorForOldMongoDB(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if client_context.version.at_least(3, 3, 9):
            return func(self, *args, **kwargs)
        else:
            with self.assertRaises(ConfigurationError):
                return func(self, *args, **kwargs)
    return wrapper


class TestCollation(unittest.TestCase):

    @classmethod
    @client_context.require_connection
    def setUpClass(cls):
        cls.listener = EventListener()
        cls.client = rs_or_single_client(event_listeners=[cls.listener])
        cls.db = cls.client.pymongo_test
        cls.collation = Collation('en_US')
        cls.warn_context = warnings.catch_warnings()
        cls.warn_context.__enter__()
        warnings.simplefilter("ignore", DeprecationWarning)

    @classmethod
    def tearDownClass(cls):
        cls.warn_context.__exit__()
        cls.warn_context = None
        cls.client.close()

    def tearDown(self):
        self.listener.results.clear()

    def last_command_started(self):
        return self.listener.results['started'][-1].command

    def assertCollationInLastCommand(self):
        self.assertEqual(
            self.collation.document,
            self.last_command_started()['collation'])

    @raisesConfigurationErrorForOldMongoDB
    def test_create_collection(self):
        self.db.test.drop()
        self.db.create_collection('test', collation=self.collation)
        self.assertCollationInLastCommand()

        # Test passing collation as a dict as well.
        self.db.test.drop()
        self.listener.results.clear()
        self.db.create_collection('test', collation=self.collation.document)
        self.assertCollationInLastCommand()

    def test_index_model(self):
        model = IndexModel([('a', 1), ('b', -1)], collation=self.collation)
        self.assertEqual(self.collation.document, model.document['collation'])

    @raisesConfigurationErrorForOldMongoDB
    def test_create_index(self):
        self.db.test.create_index('foo', collation=self.collation)
        ci_cmd = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            ci_cmd['indexes'][0]['collation'])

    @raisesConfigurationErrorForOldMongoDB
    def test_ensure_index(self):
        self.db.test.ensure_index('foo', collation=self.collation)
        ci_cmd = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            ci_cmd['indexes'][0]['collation'])

    @raisesConfigurationErrorForOldMongoDB
    def test_aggregate(self):
        self.db.test.aggregate([{'$group': {'_id': 42}}],
                               collation=self.collation)
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    @ignore_deprecations
    def test_count(self):
        self.db.test.count(collation=self.collation)
        self.assertCollationInLastCommand()

        self.listener.results.clear()
        self.db.test.find(collation=self.collation).count()
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    def test_count_documents(self):
        self.db.test.count_documents({}, collation=self.collation)
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    def test_distinct(self):
        self.db.test.distinct('foo', collation=self.collation)
        self.assertCollationInLastCommand()

        self.listener.results.clear()
        self.db.test.find(collation=self.collation).distinct('foo')
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    def test_find_command(self):
        self.db.test.insert_one({'is this thing on?': True})
        self.listener.results.clear()
        next(self.db.test.find(collation=self.collation))
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    def test_explain_command(self):
        self.listener.results.clear()
        self.db.test.find(collation=self.collation).explain()
        # The collation should be part of the explained command.
        self.assertEqual(
            self.collation.document,
            self.last_command_started()['explain']['collation'])

    @raisesConfigurationErrorForOldMongoDB
    @client_context.require_version_max(4, 1, 0, -1)
    def test_group(self):
        self.db.test.group('foo', {'foo': {'$gt': 42}}, {},
                           'function(a, b) { return a; }',
                           collation=self.collation)
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    def test_map_reduce(self):
        self.db.test.map_reduce('function() {}', 'function() {}', 'output',
                                collation=self.collation)
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    def test_delete(self):
        self.db.test.delete_one({'foo': 42}, collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['deletes'][0]['collation'])

        self.listener.results.clear()
        self.db.test.delete_many({'foo': 42}, collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['deletes'][0]['collation'])

        self.listener.results.clear()
        self.db.test.remove({'foo': 42}, collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['deletes'][0]['collation'])

    @raisesConfigurationErrorForOldMongoDB
    def test_update(self):
        self.db.test.update({'foo': 42}, {'$set': {'foo': 'bar'}},
                            collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['updates'][0]['collation'])

        self.listener.results.clear()
        self.db.test.save({'_id': 12345}, collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['updates'][0]['collation'])

        self.listener.results.clear()
        self.db.test.replace_one({'foo': 42}, {'foo': 43},
                                 collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['updates'][0]['collation'])

        self.listener.results.clear()
        self.db.test.update_one({'foo': 42}, {'$set': {'foo': 43}},
                                collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['updates'][0]['collation'])

        self.listener.results.clear()
        self.db.test.update_many({'foo': 42}, {'$set': {'foo': 43}},
                                 collation=self.collation)
        command = self.listener.results['started'][0].command
        self.assertEqual(
            self.collation.document,
            command['updates'][0]['collation'])

    @raisesConfigurationErrorForOldMongoDB
    def test_find_and(self):
        self.db.test.find_and_modify({'foo': 42}, {'$set': {'foo': 43}},
                                     collation=self.collation)
        self.assertCollationInLastCommand()

        self.listener.results.clear()
        self.db.test.find_one_and_delete({'foo': 42}, collation=self.collation)
        self.assertCollationInLastCommand()

        self.listener.results.clear()
        self.db.test.find_one_and_update({'foo': 42}, {'$set': {'foo': 43}},
                                         collation=self.collation)
        self.assertCollationInLastCommand()

        self.listener.results.clear()
        self.db.test.find_one_and_replace({'foo': 42}, {'foo': 43},
                                          collation=self.collation)
        self.assertCollationInLastCommand()

    @raisesConfigurationErrorForOldMongoDB
    def test_bulk_write(self):
        self.db.test.collection.bulk_write([
            DeleteOne({'noCollation': 42}),
            DeleteMany({'noCollation': 42}),
            DeleteOne({'foo': 42}, collation=self.collation),
            DeleteMany({'foo': 42}, collation=self.collation),
            ReplaceOne({'noCollation': 24}, {'bar': 42}),
            UpdateOne({'noCollation': 84}, {'$set': {'bar': 10}}, upsert=True),
            UpdateMany({'noCollation': 45}, {'$set': {'bar': 42}}),
            ReplaceOne({'foo': 24}, {'foo': 42}, collation=self.collation),
            UpdateOne({'foo': 84}, {'$set': {'foo': 10}}, upsert=True,
                      collation=self.collation),
            UpdateMany({'foo': 45}, {'$set': {'foo': 42}},
                       collation=self.collation)
        ])

        delete_cmd = self.listener.results['started'][0].command
        update_cmd = self.listener.results['started'][1].command

        def check_ops(ops):
            for op in ops:
                if 'noCollation' in op['q']:
                    self.assertNotIn('collation', op)
                else:
                    self.assertEqual(self.collation.document,
                                     op['collation'])

        check_ops(delete_cmd['deletes'])
        check_ops(update_cmd['updates'])

    @raisesConfigurationErrorForOldMongoDB
    def test_bulk(self):
        bulk = self.db.test.initialize_ordered_bulk_op()
        bulk.find({'noCollation': 42}).remove_one()
        bulk.find({'noCollation': 42}).remove()
        bulk.find({'foo': 42}, collation=self.collation).remove_one()
        bulk.find({'foo': 42}, collation=self.collation).remove()
        bulk.find({'noCollation': 24}).replace_one({'bar': 42})
        bulk.find({'noCollation': 84}).upsert().update_one(
            {'$set': {'foo': 10}})
        bulk.find({'noCollation': 45}).update({'$set': {'bar': 42}})
        bulk.find({'foo': 24}, collation=self.collation).replace_one(
            {'foo': 42})
        bulk.find({'foo': 84}, collation=self.collation).upsert().update_one(
            {'$set': {'foo': 10}})
        bulk.find({'foo': 45}, collation=self.collation).update({
            '$set': {'foo': 42}})
        bulk.execute()

        delete_cmd = self.listener.results['started'][0].command
        update_cmd = self.listener.results['started'][1].command

        def check_ops(ops):
            for op in ops:
                if 'noCollation' in op['q']:
                    self.assertNotIn('collation', op)
                else:
                    self.assertEqual(self.collation.document,
                                     op['collation'])

        check_ops(delete_cmd['deletes'])
        check_ops(update_cmd['updates'])

    @client_context.require_version_max(3, 3, 8)
    def test_mixed_bulk_collation(self):
        bulk = self.db.test.initialize_unordered_bulk_op()
        bulk.find({'foo': 42}).upsert().update_one(
            {'$set': {'bar': 10}})
        bulk.find({'foo': 43}, collation=self.collation).remove_one()
        with self.assertRaises(ConfigurationError):
            bulk.execute()
        self.assertIsNone(self.db.test.find_one({'foo': 42}))

    @raisesConfigurationErrorForOldMongoDB
    def test_indexes_same_keys_different_collations(self):
        self.db.test.drop()
        usa_collation = Collation('en_US')
        ja_collation = Collation('ja')
        self.db.test.create_indexes([
            IndexModel('fieldname', collation=usa_collation),
            IndexModel('fieldname', name='japanese_version',
                       collation=ja_collation),
            IndexModel('fieldname', name='simple')
        ])
        indexes = self.db.test.index_information()
        self.assertEqual(usa_collation.document['locale'],
                         indexes['fieldname_1']['collation']['locale'])
        self.assertEqual(ja_collation.document['locale'],
                         indexes['japanese_version']['collation']['locale'])
        self.assertNotIn('collation', indexes['simple'])
        self.db.test.drop_index('fieldname_1')
        indexes = self.db.test.index_information()
        self.assertIn('japanese_version', indexes)
        self.assertIn('simple', indexes)
        self.assertNotIn('fieldname', indexes)

    def test_unacknowledged_write(self):
        unacknowledged = WriteConcern(w=0)
        collection = self.db.get_collection(
            'test', write_concern=unacknowledged)
        with self.assertRaises(ConfigurationError):
            collection.update_one(
                {'hello': 'world'}, {'$set': {'hello': 'moon'}},
                collation=self.collation)
        bulk = collection.initialize_ordered_bulk_op()
        bulk.find({'hello': 'world'}, collation=self.collation).update_one(
            {'$set': {'hello': 'moon'}})
        with self.assertRaises(ConfigurationError):
            bulk.execute()
        update_one = UpdateOne({'hello': 'world'}, {'$set': {'hello': 'moon'}},
                               collation=self.collation)
        with self.assertRaises(ConfigurationError):
            collection.bulk_write([update_one])

    @raisesConfigurationErrorForOldMongoDB
    def test_cursor_collation(self):
        self.db.test.insert_one({'hello': 'world'})
        next(self.db.test.find().collation(self.collation))
        self.assertCollationInLastCommand()
